# Author: Philippe Gervais
# License: simplified BSD

# These imports are useful to launch a single test with nosetests
from . import test_class_inspect
from . import test_masking
from . import test_numpy_conversions
from . import test_datasets
from . import test_ndimage
from . import test_region
from . import test_group_sparse_covariance
from . import test_niimg_conversions
from . import test_logger
from . import test_signal
